module eposcorr_coord

  use param  
  use dal
  use sla
  use eposcorr_def


  implicit none

  private

  public correct_positions, transform_coordinates,calc_center, & 
       calc_gal_coord, coordinate_transform, meanerror

contains


  !---------------------------------------------------------------------
  subroutine correct_positions( srclist, n, ra0, dec0, &
       raoffset, decoffset, rotation)
  !---------------------------------------------------------------------
    type(Coordinate_type), dimension(:),intent(in out) :: srclist
    integer(kind=int32),intent(in) :: n
    real(kind=double),intent(in) :: ra0, dec0, raoffset, decoffset, rotation

    integer(kind=int32) :: i
    real(kind=double) :: cos_dec, sin_dx, phi, theta

    if( n <= 0 ) then
       call error('emptyList','X-ray source list does not contain data')
       return
    end if

    if( abs(raoffset) <= 1.0e-5 .AND. abs(decoffset) <= 1.0e-5 .AND. &
         abs(rotation) <= 1.0e-5 ) then
       call message(AppMsg,Verbose,'Source correction is not necessary')
       return
    end if

    sin_dx = dsin( raoffset*arcsec2rad )
    do i=1,n
       call coordinate_transform(srclist(i)%ra, srclist(i)%dec, ra0,dec0, pi,&
            srclist(i)%ra, srclist(i)%dec )

       !
       ! perform rotation
       !
       srclist(i)%ra = srclist(i)%ra + rotation*deg2rad
       !
       ! transform backward
       !
       call coordinate_transform(srclist(i)%ra, srclist(i)%dec, pi, dec0, ra0,&
            srclist(i)%ra, srclist(i)%dec )
       cos_dec = dcos( srclist(i)%dec )

       if( cos_dec > 0.0 ) then
          srclist(i)%ra = srclist(i)%ra - dasin( sin_dx/cos_dec )
       else
          call warning('invalidDec','declination is 90 degrees')
          srclist(i)%ra = 0.0
          cos_dec = 0.0
       end if
       if( srclist(i)%ra > 2.0*pi) srclist(i)%ra = srclist(i)%ra - 2.0*pi
       if( srclist(i)%ra <    0.0) srclist(i)%ra = srclist(i)%ra + 2.0*pi
       srclist(i)%dec = srclist(i)%dec - decoffset*arcsec2rad

    end do

  end subroutine correct_positions

 !---------------------------------------------------------------------
  subroutine transform_coordinates( srclist, n, ra0, dec0, invert)
 !---------------------------------------------------------------------

    type(Coordinate_type), dimension(:),intent(in out) :: srclist
    integer(kind=int32),intent(in) :: n
    real(kind=double),intent(in) :: ra0, dec0
    logical,intent(in):: invert

    integer(kind=int32) :: i
    real(kind=double),parameter :: longpole=pi

    if( n <= 0 ) then
       call error('emptyList','Source list does not contain data')
       return
    end if

    if( invert ) then
       ! inverse transform
       do i=1,n
          call coordinate_transform( srclist(i)%ra, srclist(i)%dec, longpole,&
               dec0*deg2rad, ra0*deg2rad, srclist(i)%ra, srclist(i)%dec)
       end do
    else
       do i=1,n
          call coordinate_transform(srclist(i)%ra,srclist(i)%dec, ra0*deg2rad,&
               dec0*deg2rad, longpole, srclist(i)%ra, srclist(i)%dec)
       end do
    end if

  end subroutine transform_coordinates

 !---------------------------------------------------------------------
  subroutine coordinate_transform( &
       alpha, delta, alphap, deltap, phip, phi, theta  )
 !---------------------------------------------------------------------

  ! cf Greisen & Calabretta, 1993 ,AAS Meeting, 182, #09.01
  ! Input and output value will be in radians
  
  real(kind=double),intent(in) :: alpha, delta              ! input coordinates
  real(kind=double),intent(in) :: alphap, deltap, phip      ! New poles
  real(kind=double),intent(out) :: phi, theta         ! transformed coordinates

  real(kind=double) :: alphad, deltad, alphapd, deltapd, phipd
  real(kind=double) :: phid, thetad
  real(kind=double) :: arg1, arg2

  alphad  = alpha
  deltad  = delta
  alphapd = alphap
  deltapd = deltap
  phipd   = phip 

  alphad  = alphad - alphapd

  thetad = dasin( dsin(deltad) * dsin(deltapd) + &
       dcos(deltad) * dcos(deltapd) * dcos(alphad) )
  arg1 = dsin(deltad) * dcos(deltapd) - &
       dcos(deltad) * dsin(deltapd) * dcos(alphad)
  arg2 = - dcos(deltad) * dsin(alphad)
  phid  = datan2( arg2, arg1)
  phid  =  phid + phipd

  theta = thetad
  phi   = phid

  if( phi < 0.0 ) then
     phi = phi + 2.0*pi
  end if
  if( phi > 2.0*pi ) then
     phi = phi - 2.0*pi
  end if

  end subroutine coordinate_transform


!---------------------------------------------------------------------
  subroutine calc_center( srclist, n, ra0 , dec0 )
 !---------------------------------------------------------------------
    type(Coordinate_type), dimension(:),intent(in out) :: srclist
    integer(kind=int32),intent(in) :: n
    real(kind=double),intent(out) :: ra0, dec0

    integer(kind=int32) :: i

    if( n <= 0 ) then
       call error('emptyList','Source list does not contain data')
       return
    end if

    ra0 = 0.0
    dec0 = 0.0

    do i=1,n
       ra0 = ra0 + srclist(i)%ra
       dec0 = dec0 + srclist(i)%dec 
    end do

    ra0 = ra0/n
    dec0 = dec0/n

  end subroutine calc_center



  !---------------------------------------------------------------------
  subroutine calc_gal_coord( srclist, nsrcs)
  !---------------------------------------------------------------------       
    type( Coordinate_type), dimension(:), intent(in out) :: srclist
    integer(kind=int32) , intent (in) :: nsrcs
    integer(kind=int32):: i
    real(kind=double) :: b, l

!
! use cosine formual to calculate b (NB cos(90-delta)=sin(delta)
! and use cosine + sine formual to find l

    do i=1,nsrcs
       call sla_EQGAL (srclist(i)%ra , srclist(i)%dec, l, b)
       srclist(i)%lii_corr = l*rad2deg
       srclist(i)%bii_corr = b*rad2deg
    end do

  end subroutine calc_gal_coord


  !---------------------------------------------------------------------
  function meanerror( srclist, n)
  !---------------------------------------------------------------------
    real(kind=double) :: meanerror
    type(Coordinate_type), dimension(:),intent(in) :: srclist
    integer(kind=int32),intent(in) :: n
    integer(kind=int32) :: i
    real(kind=double) :: sum, error

    meanerror = 0.0
    if( n == 0 ) return

    sum = 0.0
    do i=1,n
       error = sqrt( srclist(i)%ra_err**2 + srclist(i)%dec_err**2)
       sum = sum + error
    end do

    meanerror = sum/n

  end function meanerror


end module eposcorr_coord

