 !+
! Module: eposcorr_mod
!
! Authors:  Jacco Vink - AIP
!           Dieter Meinert - AIP (D.M.)
!          
! Description: Correction of EPIC source positions by cross-correlation with optical source catalogue.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module eposcorr_mod

  use dal
  use param  
  use errorHandling
  use eposcorr_def
  use eposcorr_io
  use eposcorr_coord
  use eposcorr_sel


  implicit none 

  private likelihood, find_offset, find_rotation, find_all, find_rot_err,&
       listparameters, printresults

  integer(kind=int16),parameter :: search_size=15, rotsize=25, coarsegridsize=5
  integer(kind=int16),parameter :: ngrid = 2*search_size+1
  real(kind=double),parameter :: min_likelihood=0.1

! introduce parameter maxposnerr D.M.20000919
! error of xray source position allowed for source to be included in 
! position fitting, dimension: arcsec
  real(kind=double) :: maxposnerr = 20.0 !ref maxdist
!

  logical           :: usebmagn, usermagn, usemaxdist, makeimage, findrotation, &
                       matchtab, nomaxsig, findlike0
  character(81)     :: xray_file, optical_file, image_file
  integer(kind=int16):: niterations=2
  integer(kind=int32):: n_matches=0
  real(kind=single) :: bmin=-10.0, bmax=30.0, rmin=-10.0, rmax=30.0
  real(kind=single) :: maxdist = 20.0
  real(kind=double) :: rawxsyserr, minxsyserr,minsys2,rawsys2
  real(kind=double) :: maximum_likelihood=0.0, total_mu=0.0, variance=0.0
  real(kind=double) :: like0
  real(kind=double) :: gridspacing=1.0, deltarota=0.2, max_error
  real(kind=double) :: maxrotation=1.0, maxoffset=8.0

  real(kind=double) :: raoffset=0.0, decoffset=0.0, rotation=0.0
  real(kind=double) :: raoffset_err=0.0, decoffset_err=0.0, rot_err=0.0
  real(kind=double) :: ra_pnt=0.0, dec_pnt=0.0, rapole, decpole

  real(kind=double),dimension(ngrid*ngrid) :: likimage

  integer(kind=int32) :: nxray=0, noptical=0, nxsum=0
  type(Coordinate_type), dimension(:), pointer :: xraylist
  type(Coordinate_type), dimension(:), pointer :: xraysummlist
  type(Coordinate_type), dimension(:), pointer :: opticallist

  character(160)    :: messagestr

  public eposcorr



contains 
  
  !---------------------------------------------------------------------
  subroutine eposcorr  
  !---------------------------------------------------------------------
    integer(kind=int16) :: i
    integer(kind=int32) :: nm
    real(kind=single) :: lik
    real(kind=double) :: averror
    character(len=20) :: xra, xdec, xerr, xext, cra,cdec,cerr,cbmg,crmg,cext
    real(kind=double) :: raoff_start,decoff_start,xref,yref
    
  
    xray_file = stringParameter('xrayset')
    optical_file = stringParameter('opticalset')
    niterations=intParameter('niter')
    findrotation = booleanParameter('findrotation')
    maxrotation = realParameter('maxrotation')
    maxoffset   = realParameter('maxoffset')
    maxdist =  realParameter('maxdist')
    usemaxdist =  booleanParameter('usemaxdist')
    usebmagn =  booleanParameter('usebmagn')
    bmin = realParameter('minbmagn')
    bmax = realParameter('maxbmagn')
    usermagn =  booleanParameter('usermagn')
    rmin = realParameter('minrmagn')
    rmax = realParameter('maxrmagn')
    makeimage = booleanParameter('makeimage')
    image_file = stringParameter('imagename')
    maxposnerr = sqrt(0.5)*realParameter('maxposnerr')
    xra = stringParameter('xrayra')
    xdec = stringParameter('xraydec')
    xerr = stringParameter ('xrayradecerr')
    xext = stringParameter ('xraytableext')
    cra = stringParameter('opticalra')
    cdec = stringParameter('opticaldec')
    cerr = stringParameter ('opticalradecerr')
    cbmg = stringParameter('bmag')
    crmg = stringParameter('rmag')
    cext = stringParameter ('opticaltableext')
    matchtab=booleanParameter("withmatchtable")
    rawxsyserr=realparameter("rawxsyserr")
    minxsyserr=realparameter("minxsyserr")
    max_error=realParameter("maxsig")
    nomaxsig=.not. booleanParameter("usemaxsig")     
  
    minsys2=2.* minxsyserr**2
    rawsys2=2.* rawxsyserr**2

    call listparameters
!
!   Read the input data sets
!
    xraylist => read_srclist( xray_file, nxray, ra_pnt, dec_pnt, &
                              xra, xdec, xerr, xext )
    opticallist => read_catset( optical_file, noptical, bmin, bmax, &
         usebmagn, rmin, rmax, usermagn, cra, cdec, cerr, cbmg, crmg, cext )
    findlike0=.true.
    if( abs(dec_pnt - nodec_pnt) < 0.1 ) then
       call warning('noRaDecPnt','Keywords ra_pnt/dec_pnt not found')
       findrotation = .FALSE.
       findlike0=.false.
    end if

!   Reduce X-ray source list to unique source positions
!    (necessary for multi-band/multi-instrument emldetect lists)

    if ( nxray > 0) xraysummlist => select_xsumm(xraylist, nxray, nxsum)

    like0=-9999.9
    if (findlike0) like0=likenull()


    if ( nxray > 0 .AND. noptical > 0 ) then

     write(errstr,*) 'Data sets were succesfully read. Processing...'
     call message(AppMsg,Verbose)

     if (booleanParameter("calculateoffsets")) then 
       
       if( usemaxdist ) then
          call message(AppMsg,Verbose,&
               'Minimizing the size of the optical data set.')
          write(errstr,*) 'Changing from  noptical = ', noptical
          call message(AppMsg,Verbose)

          call preselectlist(xraysummlist, nxsum, opticallist, noptical, maxdist )

          write(errstr,*) '           to   noptical = ', noptical
          call message(AppMsg,Verbose)
       end if

       averror = meanerror(xraysummlist, nxsum)
!       PRINT "(A F8.3)", 'average error = ', averror
       
       deltarota  = maxrotation/coarsegridsize
       gridspacing = maxoffset/coarsegridsize


       if( findrotation) then
          write(errstr,'(A,F8.3,A,F8.3)') 'Using: offset stepsize = ',&
               gridspacing,', rotation stepsize = ',deltarota
          call message(AppMsg,Verbose)

          call find_all

          call message(AppMsg,Verbose,' Results of the coarse 3D grid search:')

          call printresults
          gridspacing = 0.33*gridspacing 
          deltarota = 5.0*deltarota/(rotsize+1)
       end if

       do i=1, niterations
          write(errstr,*) 'Iteration nr ',i
          call message(AppMsg,Verbose)

          write(errstr,'(A,F8.3,A,F8.3)') 'Using: offset stepsize = ',&
               gridspacing,', rotation stepsize = ',deltarota
          call message(AppMsg,Verbose)
               
          if( i /= niterations) then
!             write(*,*) "ONLY OFFSET,grid:",gridspacing
             call find_offset
             raoffset_err = 0.0
             decoffset_err = 0.0
             if( findrotation ) call find_rotation
          else
!             write(*,*) "FIND OFFSET + FIND ROTA"
!             gridspacing = 7.0*min(averror,maxoffset)/search_size
              gridspacing = 7.0*averror/search_size
!             gridspacing = maxoffset/search_size
!             write(*,*) "gridspacing",gridspacing
             raoff_start=raoffset
             decoff_start=decoffset
             call find_offset
             if( findrotation ) then
                deltarota = 0.45*min(averror,maxrotation)/search_size
                call find_rotation
                call find_rot_err
             end if
             if( makeimage ) then
                xref=search_size+1.D0
                yref=search_size+1.D0
                call write_image(image_file, likimage, ngrid, ngrid, gridspacing, raoff_start, decoff_start, xref, yref)
                makeimage = .FALSE.
             end if
          end if

          call printresults

          if( i == niterations-1) then

          else
             gridspacing = 0.33*gridspacing 
             deltarota = 0.33*deltarota
!             write(*,*) "i == niterations-1,gridspacing = 0.33*gridspacing:", gridspacing
          end if
       end do
       

       if( n_matches <= n_min_matches .OR. &
            maximum_likelihood <= min_likelihood ) then
          call warning('invalidLikelihood', &
               'The offsets found should be considered unreliable')
       end if

       call message(AppMsg,Verbose,'Final result:')
       call printresults
       if (matchtab) lik=likelihood(raoffset,decoffset,rotation, nm, .true.)

     else
       raoffset=realParameter("raoffset")
       decoffset=realParameter("decoffset")
       rotation=realParameter("rotation")
     end if

     call correct_positions( xraylist, nxray, &
            ra_pnt*deg2rad, dec_pnt*deg2rad, raoffset, decoffset, rotation)
       

     call calc_gal_coord( xraylist, nxray)

     call modify_list(xray_file, xraylist, nxray, raoffset, decoffset, &
          rotation, raoffset_err, decoffset_err, rot_err,maximum_likelihood,&
            n_matches, like0 )
 
    else
       call error('emptyLists','One of the source lists is empty')
    end if

    if( nxray > 0 ) then
       deallocate( xraylist )
    endif

    if( noptical > 0 ) then
       deallocate( opticallist )
    endif

    if(nxsum > 0) then
      deallocate(xraysummlist)
    endif

  end subroutine eposcorr

  !---------------------------------------------------------------------
  subroutine listparameters
  !---------------------------------------------------------------------

    write(errstr,*) 'Data set with X-ray source positions :',xray_file
    call message(AppMsg, Sparse)
    write(errstr,*) 'Data set with optical source positions :',optical_file
    call message(AppMsg, Noisy)
    if( usebmagn ) then
       write(errstr,*) 'B magn range : ',bmin,' - ', bmax
       call message(AppMsg,Verbose)
    end if
    if( usermagn ) then
       write(errstr,*) 'R magn range : ',rmin,' - ', rmax
       call message(AppMsg,Verbose)
    end if
    if( usemaxdist ) then
       write(errstr,*) &
            'Maximum distance of optical to nearest X-ray source = ',maxdist
       call message(AppMsg,Verbose)
    endif
    if( makeimage ) then
       write(errstr,*) 'An image will be made with file name :', image_file
       call message(AppMsg,Verbose)
    end if
    write(errstr,'(A,F8.2)') 'maximum rotation = ',maxrotation
    call message(AppMsg,Verbose)
    write(errstr,'(A,F8.2)') 'maximum offsets  = ',maxoffset
    call message(AppMsg,Verbose)


  end subroutine listparameters

  !---------------------------------------------------------------------
  subroutine printresults
  !---------------------------------------------------------------------

       write(errstr,'(A,F8.3,A,F8.3)') ' raoffset = ',raoffset,&
            ' +/- ',raoffset_err
       call message(AppMsg,Verbose)

       write(errstr,'(A,F8.3,A,F8.3)') ' decoffset= ',decoffset,&
            ' +/- ',decoffset_err
       call message(AppMsg,Verbose)

       write(errstr,'(A,F8.3,A,F8.3)') ' rotation = ',rotation, &
            ' +/- ',rot_err
       call message(AppMsg,Verbose)

       write(errstr,'(A,F8.3,A,I4)') ' maxlik   = ',&
            maximum_likelihood,'; nm =', n_matches
       call message(AppMsg,Verbose)

       write(errstr,'(A,F8.3)') 'Likelihood expected from background sources   = ', like0
       call message(AppMsg,Verbose)
 



  end subroutine printresults

  !---------------------------------------------------------------------
  subroutine find_all
  !---------------------------------------------------------------------
    integer(kind=int16) :: i, j, k
    integer(kind=int16) :: rotsize=coarsegridsize, offsetsize=coarsegridsize
    integer(kind=int32) :: nm,nmmax
    real(kind=double)   :: best_dra, best_ddec, bestrota,maxlik
    real(kind=double)   :: delta_ra, delta_dec, rota, lik

    maxlik = maximum_likelihood
    best_dra = raoffset
    best_ddec = decoffset
    bestrota = rotation
    nmmax= n_matches

    do k = -rotsize, rotsize

       rota = rotation + k * deltarota
       do j= -offsetsize, offsetsize
          delta_dec = decoffset + j*gridspacing
          do i = -offsetsize, offsetsize
             delta_ra = raoffset + i*gridspacing
             lik =  likelihood(delta_ra, delta_dec, rota, nm, .false.)
!             write(*,*) "find_all,delta_dec,delta_ra,lik:",delta_dec,delta_ra,lik
             if ( maxlik < lik ) then
                bestrota  = rota
                best_dra  = delta_ra
                best_ddec = delta_dec
                maxlik = lik
                nmmax = nm
             end if

          end do
       end do
    end do

    maximum_likelihood = maxlik
    n_matches = nmmax
    raoffset  = best_dra
    decoffset = best_ddec
    rotation  = bestrota
    !if (matchtab) lik=likelihood(best_dra, best_ddec, bestrota, nm, .true.)

  end subroutine find_all


  !---------------------------------------------------------------------
  subroutine find_rotation
  !---------------------------------------------------------------------

    integer(kind=int16) :: i
    integer(kind=int32) :: nm, nmmax
    real(kind=double)   :: bestrota, rota, lik, maxlik



    bestrota = rotation
    maxlik = likelihood(raoffset, decoffset, rotation, nm, .false. )
    nmmax = n_matches

    do i = -rotsize, rotsize

       rota = rotation + i * deltarota
       lik =  likelihood(raoffset, decoffset, rota, nm, .false.)
       if ( maxlik < lik ) then
          bestrota = rota
          maxlik = lik
          nmmax = nm
       end if

    end do

    if( maxlik > 0.0 ) then
       rotation = bestrota
       n_matches = nmmax
       !if (matchtab) lik=likelihood(raoffset, decoffset, bestrota, nm, .true.)
    end if
    
  end subroutine find_rotation

  !---------------------------------------------------------------------
  subroutine find_offset
  !---------------------------------------------------------------------

    integer(kind=int16) :: i, j, imax, jmax, k
    integer(kind=int32) :: nm, nmmax=0
    real(kind=double) :: delta_ra, delta_dec
    real(kind=double) :: raoffset0, decoffset0
    real(kind=double) :: best_dra, best_ddec, maxlik, lik, tot_lik
    real(kind=double) :: xm, ym, rm2, r, r2, x2, y2, r_max

    raoffset0 = raoffset
    decoffset0 = decoffset


    ! search for the optimum offset using a grid
    best_dra = raoffset0
    best_ddec = decoffset0
    nmmax = n_matches
    maxlik = maximum_likelihood
    imax=0
    jmax=0
    k = 0
    do j = -search_size, search_size
       
       delta_dec = decoffset0 + gridspacing * j
       do i = -search_size, search_size
          
          delta_ra = raoffset0 + gridspacing * i
          lik =  likelihood(delta_ra, delta_dec, rotation, nm, .false.)
          ! store the likelihood values:
          k = k + 1                            ! index for the likelihood grid
          likimage( k ) = lik
          
          if ( maxlik < lik ) then
             best_dra = delta_ra
             best_ddec = delta_dec
             maxlik = lik
             nmmax = nm
             imax = i + search_size
             jmax = j + search_size
          end if

       end do
    end do

    raoffset = best_dra        
    decoffset = best_ddec
    n_matches = nmmax
    maximum_likelihood = maxlik
    !if (matchtab) lik=likelihood(best_dra,best_ddec,rotation, nm, .true.) 
    
    if( n_matches <= 0 .OR. maximum_likelihood <= 0.0 ) then
!       call warning('noMatches',&
!            'No X-ray source seems to match optical source')
       return
    endif

    r_max = 0.8*gridspacing*search_size
    xm = 0.0
    ym = 0.0
    x2 = 0.0
    y2 = 0.0
    rm2 = 0.0
    tot_lik = 0.0
    k = 0
    do j = - search_size, search_size

       delta_dec = decoffset0 + gridspacing * j
       do i = -search_size, search_size

          k = k + 1
          delta_ra = raoffset0 + gridspacing * i
          r2 = (delta_ra - best_dra)**2 +  (delta_dec - best_ddec)**2
          r = sqrt( r2 )
          if( r <= r_max ) then
             lik = likimage(k)
             xm  = xm + lik * (delta_ra - best_dra);
             ym  = ym + lik * (delta_dec - best_ddec);
             x2  = x2 + lik * (delta_ra - best_dra)**2;
             y2  = y2 + lik * (delta_dec - best_ddec)**2;
             rm2  = rm2 + lik * r2
             tot_lik = tot_lik + lik; 
          end if

       end do

    end do


    if( tot_lik > 0.0 ) then

       if( n_matches-2 > 0 ) then
          raoffset_err = sqrt(x2/(tot_lik*(n_matches-2)))
          decoffset_err = sqrt(y2/(tot_lik*(n_matches-2)))
       else
          raoffset_err = 0.0
          decoffset_err = 0.0
       endif
    end if

  end subroutine find_offset

  !---------------------------------------------------------------------
  subroutine find_rot_err
  !---------------------------------------------------------------------

    integer(kind=int16) :: i
    integer(kind=int32) :: nm
    real(kind=double)   :: rota, totlik, lik, totrot2

    totlik = 0.0
    totrot2 = 0.0

    do i = -rotsize, rotsize

       rota = rotation + i * deltarota
       lik =  likelihood(raoffset, decoffset, rota, nm, .false. )
       totrot2 = totrot2 + lik*(rota - rotation)**2
       totlik = totlik + lik
    end do

    if( totlik > 0.0 .AND. n_matches > 3) then
!       rot_err = sqrt(totrot2/(totlik - n_matches -3))
       rot_err = sqrt(totrot2/(totlik *(n_matches -3))) !!! VVH 17.06.05
    end if


  end subroutine find_rot_err

  !---------------------------------------------------------------------
  function likelihood( raoffset, decoffset, rota, nmatches, mtab)
  !---------------------------------------------------------------------

    !
    ! the likelihood is define here as:
    !                                  sum(exp( -0.5(x_i-x_j)^2/sigma^2)
    ! with x_i and x_j resp. the x-ray  and optical positions of sources 
    ! which coincide (are within 5 sigma of each other), 
    ! sigma is the position error.
    ! Note that in case the association is real and there are no systematic 
    ! errors the average likelihood is sqrt(1/2) = 0.71
    ! with a variance of sqrt(1/3)-1/2= 0.077
    !

    real(kind=double) :: likelihood
    real(kind=double), intent(in) :: raoffset, decoffset, rota
    integer(kind=int32), intent(out) :: nmatches
    logical, intent(in) :: mtab
    
    integer(kind=int32) :: i, j
    real(kind=double) :: err, xra, xdec, xra_old, xdec_old, ra0, dec0
    real(kind=double) :: err2,angdist2
    real(kind=double) :: deltaRA, angdist, cosdist, sindec, cosdec
    real(kind=double) :: rota_rad, raoffs_rad, decoffs_rad
    real(kind=double),dimension(:), allocatable, save :: xra_save, xdec_save
    real(kind=double), save :: rota_save
    logical,dimension(:), allocatable, save :: done_save
    logical, save :: first=.true.

    integer(kind=int32), dimension(:), pointer :: match1,match2
    real(kind=single), dimension(:), pointer :: match7,match8 
    real(kind=double), dimension(:), pointer :: match3,match4,match5,match6    
    integer(kind=int32), dimension(2,10000) :: imatchbuff
    real(kind=single), dimension(3:8,10000) :: matchbuff    
    real(kind=double), parameter :: del = 1d-10
    integer :: clock0,clock1,clock2,clock3

    type(DataSetT) :: set
    type(TableT) :: tab
    type(ColumnT), dimension(8) :: c



    if (first) then
      allocate(xra_save(nxsum))
      allocate(xdec_save(nxsum))
      allocate(done_save(nxsum))
      rota_save=0.d0
      done_save(:)=.false.
      first=.false.
    endif

    likelihood = 0.0
    nmatches = 0



    
    if( nxsum <= 0 .OR. noptical <= 0 ) then
       call error('emptyLists','One of the source lists does not contain data')
       return
    end if




    ra0 = ra_pnt * deg2rad
    dec0 = dec_pnt * deg2rad
    rota_rad = rota * deg2rad
    if (rota_rad /= rota_save) done_save(:)=.false.
    raoffs_rad = raoffset * arcsec2rad
    decoffs_rad = decoffset * arcsec2rad

!    call system_clock( clock0,clock2,clock3)

    do i=1, noptical
       cosdec = dcos( opticallist(i)%dec )
       sindec = dsin( opticallist(i)%dec )

       xra_old=-999d0
       xdec_old=-999d0
       do j=1,nxsum
! Add parameter maxposnerr here D.M.20000919
          if ( xraysummlist(j)%ra_err <= maxposnerr .AND. &
               xraysummlist(j)%dec_err <= maxposnerr ) then
 
            xra = xraysummlist(j)%ra
            xdec = xraysummlist(j)%dec
!           ignore additional entries of same source from emldetect lists:
!            if (xra>xra_old+del.or.xra<xra_old-del.or. &
!                xdec>xdec_old+del.or.xdec<xdec_old-del) then
!                write(*,*) "xra",xra
!             xra_old = xra
!             xdec_old = xdec 
 

!  only do coordinate transformationss, if new rotation angle, or new X-ray source (GL 05/07/2005)

             if (rota_rad == rota_save .and. done_save(j)) then
               xra=xra_save(j)
               xdec=xdec_save(j)
        
             else  
               call coordinate_transform(xra, xdec, ra0, dec0, pi, xra, xdec)
               ! perform rotation
               xra = xra +  rota_rad
               ! transform backward
               call coordinate_transform(xra, xdec, pi, dec0, ra0, xra, xdec )
               xra_save(j)=xra
               xdec_save(j)=xdec
               done_save(j)=.true.
               rota_save = rota_rad
             endif

!             err = sqrt(opticallist(i)%ra_err**2 + opticallist(i)%dec_err**2 +&
!                        xraylist(j)%ra_err**2 + xraylist(j)%dec_err**2 +&
!                        2.*rawxsyserr**2)*arcsec2rad


! simplify calculation of distances:

             err = sqrt(opticallist(i)%errsq+xraysummlist(j)%errsq+rawxsyserr)*arcsec2rad
!             err2 = (opticallist(i)%errsq+xraysummlist(j)%errsq+rawxsyserr)*arcsec2rad2*max_error**2
!             write(*,*) "err2",err2
             
             if ( err /= 0.0 ) then
                
                if (nomaxsig .or. &
                    abs(xdec-opticallist(i)%dec-decoffs_rad)/err < max_error) then

                   if( abs(cosdec) > 1.0e-8 ) then
                      deltaRA = (xra - opticallist(i)%ra - raoffs_rad/cosdec)
                   else
                      deltaRA = (xra - opticallist(i)%ra)
                   end if
                   
                  cosdist = dsin(xdec - decoffs_rad) * sindec + &
                            dcos(xdec - decoffs_rad) * cosdec * dcos(deltaRA)
                  
                   if( cosdist > 0.9 ) cosdist = min(1.0d0, cosdist) 
                   if( cosdist < -0.9 ) cosdist = max(-1.0d0, cosdist) 
                   
                   angdist = dacos(cosdist)/err             

                   
                  if(nomaxsig .or.  angdist < max_error ) then

!                      err = sqrt(opticallist(i)%ra_err**2  + &
!                                opticallist(i)%dec_err**2 + &
!                                xraylist(j)%ra_err**2 + &
!                                xraylist(j)%dec_err**2 + &
!                                2.*minxsyserr**2)*arcsec2rad
                    
                       err2 = (opticallist(i)%errsq +  xraysummlist(j)%errsq + minsys2)*arcsec2rad2

!                      write(*,*) j,xraylist(j)%ra_err,opticallist(i)%dec_err,minxsyserr, err, err**2
!                       angdist = dacos(cosdist)/err 
                       angdist2 = (dacos(cosdist))**2/err2   
!                       likelihood= likelihood+exp( -0.5*angdist*angdist)
                      likelihood= likelihood+exp( -0.5*angdist2)

                      nmatches = nmatches + 1
                      if (mtab .and. nmatches <=10000) then 
                        imatchbuff(1,nmatches)=j
                        imatchbuff(2,nmatches)=i
                        matchbuff(3,nmatches)=xra/deg2rad
                        matchbuff(4,nmatches)=xdec/deg2rad
                        matchbuff(5,nmatches)=opticallist(i)%ra/deg2rad
                        matchbuff(6,nmatches)=opticallist(i)%dec/deg2rad
                        matchbuff(7,nmatches)=dacos(cosdist)/deg2rad*3600.
                        matchbuff(8,nmatches)=exp(-0.5*angdist2)
                      end if 
                   end if
                   
                end if
!              end if
            end if
          end if ! maxposnerr
       end do ! nxsum
    end do

!    write(*,*) raoffset,decoffset,rota,likelihood

    if (mtab) then
      if (nmatches>10000) call warning("BufferFull", &
         "Buffer full; only first 10000 matches listed")

      set = dataSet(stringParameter("matchset"),CREATE)
      tab = addTable (set, "SRCLIST", nmatches)
      c(1)=addColumn(tab,"X_INDEX",Integer32," ")
      match1 => int32Data(c(1))
      match1 = imatchbuff(1,1:nmatches)
      c(3)=addColumn(tab,"X_RA",Real64,"deg")
      match3 => real64Data(c(3))
      match3 = matchbuff(3,1:nmatches)
      c(4)=addColumn(tab,"X_DEC",Real64,"deg")
      match4 => real64Data(c(4))
      match4 = matchbuff(4,1:nmatches)
      c(2)=addColumn(tab,"O_INDEX",Integer32," ")
      match2 => int32Data(c(2))
      match2 = imatchbuff(2,1:nmatches)
      c(5)=addColumn(tab,"O_RA",Real64,"deg")
      match5 => real64Data(c(5))
      match5 = matchbuff(5,1:nmatches)
      c(6)=addColumn(tab,"O_DEC",real64,"deg")
      match6 => real64Data(c(6))
      match6 = matchbuff(6,1:nmatches)
      c(7)=addColumn(tab,"DIST",real32,"arcsec")
      match7 => real32Data(c(7))
      match7 = matchbuff(7,1:nmatches)
      c(8)=addColumn(tab,"LIKELIHOOD",real32," ")
      match8 => real32Data(c(8))
      match8 = matchbuff(8,1:nmatches)
      do i=1,8
        call release(c(i))
      end do
      call release(tab)
      call release(set)

    end if

!  call system_clock( clock1,clock2,clock3)
!  write(*,*) "CPU time (likelihood):",clock1-clock0

  end function likelihood


  !---------------------------------------------------------------------  
  function likenull() 
  !---------------------------------------------------------------------  

    ! Calculate null hypothesis likelihood for the assumption that all matches are
    ! spurious. The calculation is based on the density of backgrounds sources n.
    ! Each X-ray source produces a likelihood due to spurious matches 
    ! l0=2*pi*sigma**2*n.
    ! The density n is calculated within a radius of 10 arcmin from (ra_pnt,dec_pnt).
    ! The calculation is only correct, if the total number of optical sources >>
    ! number of X-ray sources. likenull is meaningless, if only likely counterparts
    ! are in the "optical" list, e.g. if 2 X-ray source lists are correlated with
    ! each other. 

    real(kind=double) :: likenull
    integer (kind=int32) :: i, j, ncount
    real(kind=double) :: errsq, maxangdist, maxangdeg=0.1666666667
    real(kind=double) :: ra0, dec0, cosdec, angdist, xra_old, xdec_old
    real(kind=single) :: density,errsqmean
    real(kind=double), parameter :: del = 1d-10



    ! calculate optical source density:
    maxangdist=maxangdeg*deg2rad
    ncount=0
    ra0 = ra_pnt * deg2rad
    dec0 = dec_pnt * deg2rad
    cosdec =  dcos(dec0)
    do i=1, noptical
      angdist=sqrt((opticallist(i)%dec-dec0)**2+((opticallist(i)%ra-ra0)*cosdec)**2)
      if (angdist <= maxangdist) ncount=ncount+1
    enddo
  
    density=float(ncount)/(pi*(maxangdeg*3600.)**2)
    ! calculate average of optical squared errors.
    ! Using the average error  is an approximation, but in most cases 
    ! the optical sources will have identical errors anyway:
    errsqmean=0.0
    do i=1,noptical
       errsqmean=errsqmean+opticallist(i)%errsq
    enddo
    errsqmean=errsqmean/float(noptical)

    ! sum background contributions from all X-ray sources:
    likenull=0.0
!    xra_old=-999d0
!    xdec_old=-999d0
    do j=1,nxsum
!      if (abs(xraylist(j)%ra-xra_old) > del .or. abs(xraysummlist(j)%dec-xdec_old) > del) then
        if ( xraysummlist(j)%ra_err <= maxposnerr .AND. &
                 xraysummlist(j)%dec_err <= maxposnerr ) then
          errsq = xraysummlist(j)%errsq+errsqmean+minsys2
          likenull=likenull+2*pi*errsq*density
        endif
!        xra_old=xraylist(j)%ra
!        xdec_old=xraylist(j)%dec
!      endif 
    enddo

   
  end function likenull


end module eposcorr_mod
